#include <pty.h>
#include <utmp.h>
#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <event.h>
#include <fcntl.h>
#include <termios.h>
#include <openssl/md5.h>
#include <string.h>
#include <assert.h>
#include <errno.h>
#include <sys/signal.h>
#include <sys/ioctl.h>

#define MAX_BUF_SIZE 16
#define MAX_MD5_SUMS 16

static int master;
static struct event ev_read[2];
static struct event ev_sig;
static char buffer[MAX_BUF_SIZE] = { 0 };
static int buf_at = 0;
//unsigned char prevent_md5[16] = { 0x91, 0x2e, 0xc8, 0x03, 0xb2, 0xce, 0x49, 0xe4, 0xa5, 0x41, 0x06, 0x8d, 0x49, 0x5a, 0xb5, 0x70 };
//unsigned char prevent_md5[16] = { 0x0c, 0xc1, 0x75, 0xb9, 0xc0, 0xf1, 0xb6, 0xa8, 0x31, 0xc3, 0x99, 0xe2, 0x69, 0x77, 0x26, 0x61  };
//unsigned char prevent_md5[16]= { 0x5d, 0x41, 0x40, 0x2a, 0xbc, 0x4b, 0x2a, 0x76, 0xb9, 0x71, 0x9d, 0x91, 0x10, 0x17, 0xc5, 0x92 };
static unsigned char prevent_md5[MAX_MD5_SUMS][16];
static int prevent_md5_n = 0;
int scrubbing = 0;

void
handle_sig(int a_sig, short a_event, void* a_arg)
{
  struct winsize size;
  if (ioctl(0, TIOCGWINSZ, (char *)&size) == 0) {
    ioctl(master, TIOCSWINSZ, (char *)&size);
  }
  signal_add(&ev_sig, NULL);
}

void
handle_read(int a_fd, short a_event, void* a_arg)
{
  char buf[1024];
  int n;
  int from, to;
  MD5_CTX md5;
  int fail = 0;
  if (a_fd == 0) {
    event_add(&ev_read[0], NULL);
    from = 0;
    to = master;
  } else {
    event_add(&ev_read[1], NULL);
    from = master;
    to = 1;
  }
  n = read(from, buf, 1024);
  if (n > 0) {
    if (a_fd == 0) {
      int c;
      int i;
      if (scrubbing) {
        for (i = 0; i < n && buf[i] != '\r'; ++i) {
          // iterate
        }
        if (i < n && buf[i] == '\r') {
          scrubbing = 0;
          ++i;
        }
        memmove(buf + i, buf, n - i);
        n -= i;
      }
      for (i = 0; i < n && !scrubbing; ++i) {
        int len;
        buffer[buf_at] = buf[i];
        buf_at = (buf_at + 1) % MAX_BUF_SIZE;
        for (len = 1; len < MAX_BUF_SIZE && !scrubbing; ++len) {
          char md[16];
          MD5_Init(&md5);
          int at = buf_at - len;
          while (at < 0) {
            at += MAX_BUF_SIZE;
          }
          for (c = 0; c < len; ++c) {
            MD5_Update(&md5, buffer + at, 1);
            at = (at + 1) % MAX_BUF_SIZE;
          }
          MD5_Final(md, &md5);

          int any_match = 0;
          for (c = 0; c < prevent_md5_n && !any_match; ++c) {
            any_match = (memcmp(md, prevent_md5[c], 16) == 0);
          }
          if (any_match) {
            scrubbing = 1;
            // discard the rest of the buffer, minus the letters that were written
            n = i - (len - 1);
            while (n < 0) {
              char ch = 0x08;
              char erase[] = { 0x08, ' ', 0x08 };
              write(to, &ch, 1);
              write(1, erase, 3);
              ++n;
            }
            printf("SCRUBBING");
            fflush(stdout);
          }
        }
      }
    }
    {
      int ofs = 0;
      while (ofs < n && !fail) {
        int wrote = write(to, buf + ofs, n - ofs);
        if (wrote > 0) {
          ofs += wrote;
        } else if (wrote == 0 || (wrote < 0 && errno != EAGAIN && errno != EINTR)) {
          fail = 1;
        }
      }
    }
  } else {
    fail = 1;
  }
  if (fail) {
    signal_del(&ev_sig);
    event_del(&ev_read[1]);
    event_del(&ev_read[0]);
  }
}

int
main(int a_c, char** a_v)
{
  struct winsize size;
  pid_t pid;
  {
    int opt;
    while ((opt = getopt(a_c, a_v, "s:")) != -1) {
      switch (opt) {
      case 's': {
        const char hexchars[] = "0123456789abcdef";
        int c;
        assert(prevent_md5_n < MAX_MD5_SUMS);
        assert(strlen(hexchars) == 16);
        assert(strlen(optarg) == 32);
        for (c = 0; c < 16; ++c) {
          char* high = strchr(hexchars, optarg[c * 2]);
          char* low = strchr(hexchars, optarg[c * 2 + 1]);
          assert(high);
          assert(low);
          prevent_md5[prevent_md5_n][c] = (unsigned char)(high - hexchars) * 0x10 + (unsigned char)(low - hexchars);
        }
        ++prevent_md5_n;
      } break;
      }
    }
  }

  ioctl(0, TIOCGWINSZ, &size);
  pid = forkpty(&master, NULL, NULL, &size);
  struct termios orig_tt;
  tcgetattr(0, &orig_tt);

  if (!pid) {
    if (optind < a_c) {
      execvp(a_v[optind], a_v + optind);
    } else {
      execl(getenv("SHELL"), getenv("SHELL"), NULL);
    }
    _exit(1);
  }

  fcntl(0, O_NONBLOCK, 1);
  fcntl(master, O_NONBLOCK, 1);

  {
    struct termios t = orig_tt;
    cfmakeraw(&t);
    t.c_lflag &= ~ECHO;
    tcsetattr(0, TCSAFLUSH, &t);
  }

  event_init();
  event_set(&ev_read[0], 0, EV_READ, handle_read, NULL);
  event_set(&ev_read[1], master, EV_READ, handle_read, NULL);
  signal_set(&ev_sig, SIGWINCH, handle_sig, NULL);
  event_add(&ev_read[0], NULL);
  event_add(&ev_read[1], NULL);
  signal_add(&ev_sig, NULL);
  event_dispatch();
  
  tcsetattr(0, TCSAFLUSH, &orig_tt);
}
